/*
 * Hardware monitoring driver for Texas Instruments TPS53679
 *
 * Copyright (c) 2017 Mellanox Technologies. All rights reserved.
 * Copyright (c) 2017 Vadim Pasternak <vadimp@mellanox.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/err.h>
#include <linux/i2c.h>
#include "pmbus.h"

#define TPS53679_PROT_VR12_5MV		0x01 /* VR12.0 mode, 5-mV DAC */
#define TPS53679_PROT_VR12_5_10MV	0x02 /* VR12.5 mode, 10-mV DAC */
#define TPS53679_PROT_VR13_10MV		0x04 /* VR13.0 mode, 10-mV DAC */
#define TPS53679_PROT_IMVP8_5MV		0x05 /* IMVP8 mode, 5-mV DAC */
#define TPS53679_PROT_VR13_5MV		0x07 /* VR13.0 mode, 5-mV DAC */
#define TPS53679_PAGE_NUM		2
#define TPS53679_VOUT_OFFSET_VR12	250
#define TPS53679_VOUT_FACTOR_VR12	5
#define TPS53679_VOUT_OFFSET_VR13	(TPS53679_VOUT_OFFSET_VR12 * 2)
#define TPS53679_VOUT_FACTOR_VR13	(TPS53679_VOUT_FACTOR_VR12 * 2)

#define TPS53679_VOUT_REG2VAL(off, val, factor, scale) \
	(((off) + ((val) - 1) * (factor)) * (scale))

#define TPS53679_VOUT_VAL2REG(off, rv, factor, scale) \
	(DIV_ROUND_CLOSEST(DIV_ROUND_CLOSEST(rv, (scale)) - (off), (factor)) \
	+ 1)

struct tps53679_data {
	u8 vout_scale[TPS53679_PAGE_NUM];	/* VOUT scale per page */
	u8 vout_params;				/* VOUT parameters */
	struct pmbus_driver_info info;
};

#define to_tps53679_data(x)  container_of(x, struct tps53679_data, info)

static int tps53679_identify(struct i2c_client *client,
			     struct pmbus_driver_info *info)
{
	struct tps53679_data *data = to_tps53679_data(info);
	int ret;

	/* Read custom register with VOUT scaling value.*/
	ret = pmbus_read_byte_data(client, 0, PMBUS_VOUT_MODE);
	if (ret < 0)
		return ret;

	data->vout_params = ret & GENMASK(4, 0);
	switch (data->vout_params) {
	case TPS53679_PROT_VR13_10MV:
	case TPS53679_PROT_VR12_5_10MV:
		info->vrm_version = vr13;
		break;
	case TPS53679_PROT_VR13_5MV:
	case TPS53679_PROT_VR12_5MV:
	case TPS53679_PROT_IMVP8_5MV:
		info->vrm_version = vr12;
		break;
	default:
		return -EINVAL;
	}

	return 0;
}

static int
tps53679_read_word_data(struct i2c_client *client, int page, int reg)
{
	const struct pmbus_driver_info *info = pmbus_get_driver_info(client);
	struct tps53679_data *data = to_tps53679_data(info);
	long rv;
	int ret;

	switch (reg) {
	case PMBUS_READ_VOUT:
		ret = pmbus_read_word_data(client, page, PMBUS_READ_VOUT);
		if (ret < 0)
			return ret;

		if (!data->vout_scale[page])
			break;

		/* Value should be scaled, before it presented to sysfs. */
		switch (data->vout_params) {
		case TPS53679_PROT_VR13_10MV:
		case TPS53679_PROT_VR12_5_10MV:
			rv = TPS53679_VOUT_REG2VAL(TPS53679_VOUT_OFFSET_VR13,
						   ret,
						   TPS53679_VOUT_FACTOR_VR13,
						   data->vout_scale[page]);
			ret = TPS53679_VOUT_VAL2REG(TPS53679_VOUT_OFFSET_VR13,
						    rv,
						   TPS53679_VOUT_FACTOR_VR13,
						   data->vout_scale[page]);
			break;

		case TPS53679_PROT_VR13_5MV:
		case TPS53679_PROT_VR12_5MV:
		case TPS53679_PROT_IMVP8_5MV:
			rv = TPS53679_VOUT_REG2VAL(TPS53679_VOUT_OFFSET_VR12,
						   ret,
						   TPS53679_VOUT_FACTOR_VR12,
						   data->vout_scale[page]);
			ret = TPS53679_VOUT_VAL2REG(TPS53679_VOUT_OFFSET_VR12,
						    rv,
						   TPS53679_VOUT_FACTOR_VR12,
						   data->vout_scale[page]);
			break;

		default:
			return -EINVAL;
	}

		break;
	default:
		return -ENODATA;
	}
	return ret;
}

static struct pmbus_driver_info tps53679_info = {
	.pages = TPS53679_PAGE_NUM,
	.format[PSC_VOLTAGE_IN] = linear,
	.format[PSC_VOLTAGE_OUT] = vid,
	.format[PSC_TEMPERATURE] = linear,
	.format[PSC_CURRENT_OUT] = linear,
	.format[PSC_POWER] = linear,
	.func[0] = PMBUS_HAVE_VIN | PMBUS_HAVE_VOUT | PMBUS_HAVE_STATUS_VOUT |
		PMBUS_HAVE_IOUT | PMBUS_HAVE_STATUS_IOUT |
		PMBUS_HAVE_TEMP | PMBUS_HAVE_STATUS_TEMP |
		PMBUS_HAVE_POUT,
	.func[1] = PMBUS_HAVE_VIN | PMBUS_HAVE_VOUT | PMBUS_HAVE_STATUS_VOUT |
		PMBUS_HAVE_IOUT | PMBUS_HAVE_STATUS_IOUT |
		PMBUS_HAVE_TEMP | PMBUS_HAVE_STATUS_TEMP |
		PMBUS_HAVE_POUT,
	.read_word_data = tps53679_read_word_data,
	.identify = tps53679_identify,
};

static int tps53679_probe(struct i2c_client *client,
			  const struct i2c_device_id *id)
{
	struct tps53679_data *data;

	data = devm_kzalloc(&client->dev, sizeof(*data), GFP_KERNEL);
	if (!data)
		return -ENOMEM;

	return pmbus_do_probe(client, id, &tps53679_info);
}

static const struct i2c_device_id tps53679_id[] = {
	{"tps53679", TPS53679_PAGE_NUM},
	{}
};

MODULE_DEVICE_TABLE(i2c, tps53679_id);

static const struct of_device_id tps53679_of_match[] = {
	{.compatible = "ti,tps53679",},
};
MODULE_DEVICE_TABLE(of, tps53679_of_match);

static struct i2c_driver tps53679_driver = {
	.driver = {
		.name = "tps53679",
		.of_match_table = of_match_ptr(tps53679_of_match),
	},
	.probe = tps53679_probe,
	.remove = pmbus_do_remove,
	.id_table = tps53679_id,
};

module_i2c_driver(tps53679_driver);

MODULE_AUTHOR("Vadim Pasternak <vadimp@mellanox.com>");
MODULE_DESCRIPTION("PMBus driver for Texas Instruments TPS53679");
MODULE_LICENSE("GPL");
